% Example of Harris corner detection
%
% Christian Forster

% Harris corner detector according to:
% C. Harris and M. Stephens, "A combined corner and edge detector," 
% in Proceedings of The Fourth Alvey Vision Conference, pp. 147-151, 1988. 

% You can run the entire script by pressing the F5 key
% Or, you can evaluate one or more lines by selecting them and pressing the
% F9 key (without needing to copy them to the prompt in the Command Window)

clear all % clear all variables in the workspace
close all % close all figures
clc       % clear the command window


% Load stereo images
I_left = double(imread('stereo_left.png'))/255;
I_right = double(imread('stereo_right.png'))/255;

% Params
display_intermediate_results = false;


%% Extract harris corners in both images

corners_left = harrisCorners(I_left, 5, 0.04, 0.05, display_intermediate_results);
corners_right = harrisCorners(I_right, 5, 0.04, 0.05, display_intermediate_results);

figure(5)
subplot(1,2,1)
imshow(I_left, [])
hold on
plot(corners_left(1,:), corners_left(2,:) ,'ys')
title('left stereo image')

subplot(1,2,2)
imshow(I_right, [])
hold on
plot(corners_right(1,:), corners_right(2,:) ,'ys')
title('right stereo image')


%% Basic feature matching using Sum of Squared Differences (SSD) score

num_corners_left = size(corners_left, 2);
num_corners_right = size(corners_right, 2);
[height, width] = size(I_left);
halfpatch_size = 5;
display_intermediate_results = false;
matches = zeros(num_corners_left, 2);

% loop through all corners in the left image
for i_left = 1:num_corners_left
    
    x_left = corners_left(1, i_left);
    y_left = corners_left(2, i_left);
    
    % check if left corner is not too close to the border
    if x_left-halfpatch_size >= 1 && y_left-halfpatch_size >= 1 ...
        && x_left+halfpatch_size < width && y_left+halfpatch_size < height
        
       left_patch = I_left(y_left-halfpatch_size:y_left+halfpatch_size, ...
                           x_left-halfpatch_size:x_left+halfpatch_size);
       
       if display_intermediate_results
           figure(2)
           imshow(left_patch, [])
           waitforbuttonpress
       end
       
       % now loop through all features in the right image and find best
       % match
       i_right_best = 0;
       ssd_best = 100000000000;
       for i_right = 1:num_corners_right
           
           x_right = corners_right(1, i_right);
           y_right = corners_right(2, i_right);

           % check if left corner is not too close to the border
           if x_right-halfpatch_size >= 1 && y_right-halfpatch_size >= 1 ...
                && x_right+halfpatch_size < width && y_right+halfpatch_size < height
           
               right_patch = I_right(y_right-halfpatch_size:y_right+halfpatch_size, ...
                                     x_right-halfpatch_size:x_right+halfpatch_size);
                                 
               % compute SSD
               ssd = sum(sum((right_patch-left_patch).^2));
               if ssd < ssd_best
                   i_right_best = i_right;
                   ssd_best = ssd;
               end
           end
       end
       
       % save best match
       matches(i_left,:) = [i_right_best, ssd_best];
    end
end

%% display matches

% first, combine the two images in one large image
I_combined = [I_left, I_right];

figure()
imshow(I_combined, [])
hold on

% draw matches
ssd_threshold = 1.0
for i = 1:size(matches,1)
    if matches(i,1) > 0 && matches(i,2) < ssd_threshold
        pt_left = corners_left(:,i);
        pt_right = corners_right(:,matches(i,1)) + [width; 0];
        color = rand(1,3);
        plot([pt_left(1), pt_right(1)], [pt_left(2), pt_right(2)], '-', 'Color', color);
    end
end
